#========================================================================
# Package:  html-ext.pl, for j?weblint 97, v0.12 [1997-03-08]
# Purpose:  Add HTML extensions to j?weblint 97.
#           The default spec of this version is *HTML 2.0* (RFC 1866).
#           Currently supported extensions are:
#             HTML3.2    - HTML 3.2 extensions
#                         See: http://www.w3.org/pub/WWW/MarkUp/Wilbur/
#                         Note: HTML 3.2 has become a W3C Recommendation
#                               on Jan. 14, 1996.
#             i18n      - extended for i18n, specified in RFC 2070:
#                         "Internationalization of the Hypertext Markup
#                          Language". 
#             Cougar    - EXPERIMENTAL extensions to HTML 3.2
#                         (based on 1996-07-12 draft). See: 
#                         http://www.w3.org/pub/WWW/MarkUp/Cougar/HTML.dtd
#             Microsoft - the extensions for Microsoft Internet Explorer 3.0,
#                         based on 1996-11-07 DTD for IE 3.0. See: 
#                     http://www.microsoft.com/workshop/author/ref/ie3dtd-f.htm
#                         and IE 3.0 HTML Reference:
#                 http://www.microsoft.com/workshop/author/newhtml/htmlr020.htm
#             Netscape  - the extensions for Netscape Navigator 3.0,
#                         based on Netscape HTML Reference Guide:
#        http://developer.netscape.com/library/documentation/htmlguid/index.htm
#========================================================================

#========================================================================
# DOCTYPE declarations
#========================================================================
$HTML2DocType  = 'HTML\s+PUBLIC\s+["\x27](HTML|-//IETF//DTD\s+HTML(\s+Strict)?(//EN(2\.0)?|\s+2.0(\s+Strict)?(\s+Level\s+2)?//EN|(\s+Strict)?\s+Level\s+2//EN(//2\.0)?))["\x27]';
$WilburDocType ='HTML\s+PUBLIC\s+["\\x27]-//W3C//DTD\s+HTML\s+3\.2\s*(Final|Draft|Draft\s+19960821)?//EN["\\x27]';
$i18nDocType   = 'HTML\s+PUBLIC\s+["\x27]-//IETF//DTD\s+HTML\s+i18n//EN["\x27]';
$CougarDocType = 'HTML\s+PUBLIC\s+["\x27]-//W3C//DTD\s+HTML\s+Experimental\s+199607(09|1[02])//EN["\x27]';
$msDocType     = 'HTML\s+PUBLIC\s+["\x27]-//Microsoft//DTD\s+Internet\s+Explorer\s*[23].0.*//EN["\x27]';

#========================================================================
# HTML 3.2 (Wilbur) elements and attributes
#========================================================================

$WilburElements = 'APPLET|AREA|BASEFONT|BIG|CAPTION|CENTER|DIV|FONT|MAP|'.
                  'PARAM|SCRIPT|SMALL|STRIKE|STYLE|SUB|SUP|TABLE|TD|TH|TR|U';

%WilburAttributes =
(
 'APPLET',   'ALIGN|ALT|CODE|CODEBASE|HEIGHT|HSPACE|NAME|VSPACE|WIDTH',
 'AREA',     'ALT|COORDS|HREF|NOHREF|SHAPE',
 'BASEFONT', 'SIZE',
 'BODY',     'ALINK|BACKGROUND|BGCOLOR|LINK|TEXT|VLINK',
 'BR',       'CLEAR',
 'CAPTION',  'ALIGN',
 'DIR',      'COMPACT',
 'DIV',      'ALIGN',
 'FONT',     'COLOR|SIZE',
 'H1',       'ALIGN',
 'H2',       'ALIGN',
 'H3',       'ALIGN',
 'H4',       'ALIGN',
 'H5',       'ALIGN',
 'H6',       'ALIGN',
 'HR',       'ALIGN|NOSHADE|SIZE|WIDTH',
 'IMG',      'BORDER|HEIGHT|HSPACE|USEMAP|VSPACE|WIDTH',
 'ISINDEX',  'PROMPT',
 'LI',       'TYPE|VALUE',
 'MAP',      'NAME',
 'MENU',     'COMPACT',
 'OL',       'START|TYPE',
 'P',        'ALIGN',
 'PARAM',    'NAME|VALUE',
 'TABLE',    'ALIGN|BORDER|CELLPADDING|CELLSPACING|WIDTH',
 'TD',       'ALIGN|COLSPAN|HEIGHT|NOWRAP|ROWSPAN|VALIGN|WIDTH',
 'TH',       'ALIGN|COLSPAN|HEIGHT|NOWRAP|ROWSPAN|VALIGN|WIDTH',
 'TR',       'ALIGN|VALIGN',
 'UL',       'TYPE',
);

$colorRE = '#[0-9a-fA-F]{6}';
$ColorNames = 'aqua|black|blue|fuchsia|gray|green|lime|maroon|'.
              'navy|olive|purple|red|silver|teal|white|yellow';

#========================================================================
# i18n elements
#========================================================================

$i18nEntities = 'zwnj|zwj|lrm|rlm';

$i18nElements = 'BDO|Q|SPAN|SUB|SUP';

%i18nAttributes =
(
 'A',          'CHARSET|CLASS|DIR|ID|LANG',
 'ADDRESS',    'ALIGN|CLASS|DIR|ID|LANG',
 'BDO',        'CLASS|DIR|ID|LANG',
 'BLOCKQUOTE', 'ALIGN|CLASS|DIR|ID|LANG',
 'DIR',        'ALIGN|CLASS|DIR|ID|LANG',
 'DL',         'CLASS|DIR|ID|LANG',
 'FORM',       'CLASS|DIR|ID|LANG',
 'H1',         'ALIGN|CLASS|DIR|ID|LANG',
 'H2',         'ALIGN|CLASS|DIR|ID|LANG',
 'H3',         'ALIGN|CLASS|DIR|ID|LANG',
 'H4',         'ALIGN|CLASS|DIR|ID|LANG',
 'H5',         'ALIGN|CLASS|DIR|ID|LANG',
 'H6',         'ALIGN|CLASS|DIR|ID|LANG',
 'HR',         'ALIGN|CLASS|DIR|ID|LANG',
 'HTML',       'CLASS|DIR|ID|LANG',
 'IMG',        'CLASS|DIR|ID|LANG',
 'INPUT',      'ACCEPT|ACCEPT-CHARSET|CLASS|DIR|ID|LANG',
 'LI',         'ALIGN|CLASS|DIR|ID|LANG',
 'LINK',       'CHARSET|CLASS|DIR|ID|LANG',
 'MENU',       'ALIGN|CLASS|DIR|ID|LANG',
 'OL',         'ALIGN|CLASS|DIR|ID|LANG',
 'OPTION',     'CLASS|DIR|ID|LANG',
 'P',          'ALIGN|CLASS|DIR|ID|LANG',
 'PRE',        'CLASS|DIR|ID|LANG',
 'TEXTAREA',   'ACCEPT-CHARSET|CLASS|DIR|ID|LANG',
 'UL',         'ALIGN|CLASS|DIR|ID|LANG',
);

#========================================================================
# Cougar elements and attributes
#========================================================================

$CougarEntities = 'trade|thinsp|emsp|ensp|mdash|ndash|zwnj|zwj|lrm|rlm';

$CougarElements = 'APPLET|AREA|BASEFONT|BDO|BIG|CAPTION|CENTER|COL|COLGROUP|'.
                  'DIV|FIELDSET|FONT|LABEL|MAP|OBJECT|PARAM|Q|S|SCRIPT|SMALL|'.
                  'SPAN|STYLE|SUB|SUP|TABLE|TBODY|TD|TEXTFLOW|TFOOT|TH|THEAD|'.
                  'TR|U';

%CougarAttributes =
(
 'A',          'ACCESSKEY|CLASS|COORDS|DIR|ID|NOTAB|LANG|ONCLICK|ONMOUSEOUT|'.
               'ONMOUSEOVER|SHAPE|STYLE|TABINDEX',
 'APPLET',     'ALIGN|ALT|CLASS|CODE|CODEBASE|DOWNLOAD|HEIGHT|HSPACE|ID|NAME|'.
               'STYLE|TITLE|VSPACE|WIDTH',
 'AREA',       'ALT|COORDS|HREF|NOHREF|NOTAB|ONCLICK|ONMOUSEOUT|ONMOUSEOVER|'.
               'SHAPE|TABINDEX|TITLE',
 'BASEFONT',   'SIZE',
 'BDO',        'DIR|LANG',
 'BLOCKQUOTE', 'CLASS|DIR|ID|LANG|STYLE',
 'BODY',       'ALINK|BACKGROUND|BGCOLOR|CLASS|DIR|ID|LANG|LINK|'.
               'ONLOAD|ONUNLOAD|STYLE|TEXT|VLINK',
 'BR',         'CLASS|CLEAR|DIR|ID|LANG|STYLE',
 'CAPTION',    'ACCESSKEY|ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'COL',        'ALIGN|CHAR|CHAROFF|CLASS|DIR|ID|LANG|SPAN|STYLE|VALIGN|WIDTH',
 'COLGROUP',   'ALIGN|CHAR|CHAROFF|CLASS|DIR|ID|LANG|SPAN|STYLE|VALIGN|WIDTH',
 'DD',         'CLASS|DIR|ID|LANG|STYLE',
 'DIR',        'COMPACT',
 'DIV',        'ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'DL',         'CLASS|DIR|ID|LANG|STYLE',
 'DT',         'CLASS|DIR|ID|LANG|STYLE',
 'FIELDSET',   'CLASS|DIR|ID|LANG|STYLE|TITLE',
 'FONT',       'COLOR|FACE|SIZE',
 'FORM',       'ONSUBMIT',
 'H1',         'ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'H2',         'ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'H3',         'ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'H4',         'ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'H5',         'ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'H6',         'ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'HR',         'ALIGN|CLASS|ID|NOSHADE|SIZE|STYLE|WIDTH',
 'IMG',        'BORDER|CLASS|HEIGHT|HSPACE|ID|STYLE|TITLE|USEMAP|VSPACE|WIDTH',
 'INPUT',      'CLASS|DIR|ID|LANG|NOTAB|ONBLUR|ONCHANGE|ONCLICK|ONFOCUS|'.
               'ONSELECT|STYLE|TABINDEX|TITLE',
 'ISINDEX',    'DIR|LANG|PROMPT',
 'LABEL',      'ACCESSLKEY|CLASS|DIR|FOR|ID|LANG|ONCLICK|STYLE|TITLE',
 'LI',         'CLASS|DIR|ID|LANG|STYLE|TYPE|VALUE',
 'LINK',       'CLASS|ID|STYLE|TYPE',
 'MAP',        'NAME',
 'MENU',       'COMPACT',
 'OBJECT',     'ALIGN|BORDER|CLASS|CLASSID|CODEBASE|CODETYPE|DATA|DECLARE|'.
               'DIR|HEIGHT|HSPACE|ID|LANG|NAME|NOTAB|SHAPES|STANDBY|STYLE|'.
               'TABINDEX|TITLE|TYPE|USEMAP|VSPACE|WIDTH|',
 'OL',         'CLASS|DIR|ID|LANG|START|STYLE|TYPE',
 'OPTION',     'CLASS|DIR|ID|LANG|STYLE',
 'P',          'ALIGN|CLASS|DIR|ID|LANG|STYLE',
 'PARAM',      'NAME|TYPE|VALUE|VALUETYPE',
 'PRE',        'CLASS|DIR|ID|LANG|STYLE',
 'SCRIPT',     'LANGUAGE|SRC|TYPE',
 'SELECT',     'CLASS|DIR|ID|LANG|NOTAB|ONBLUR|ONCHANGE|ONFOCUS|STYLE|'.
               'TABINDEX|TITLE',
 'SPAN',       'CLASS|DIR|ID|LANG|STYLE',
 'STYLE',      'DIR|LANG|TITLE|TYPE',
 'TABLE',      'ALIGN|BORDER|CELLPADDING|CELLSPACING|CLASS|COLS|DIR|FRAME|ID|'.
               'LANG|RULES|STYLE|WIDTH',
 'TBODY',      'ALIGN|CHAR|CHAROFF|CLASS|DIR|ID|LANG|STYLE|VALIGN',
 'TD',         'ALIGN|AXES|AXIS|BGCOLOR|CHAR|CHAROFF|CLASS|COLSPAN|DIR|ID|'.
               'LANG|NOWRAP|ROWSPAN|STYLE|VALIGN',
 'TEXTAREA',   'CLASS|DIR|ID|LANG|NOTAB|ONBLUR|ONCHANGE|ONFOCUS|ONSELECT|'.
               'STYLE|TABINDEX|TITLE',
 'TFOOT',      'ALIGN|CHAR|CHAROFF|CLASS|DIR|ID|LANG|STYLE|VALIGN',
 'TH',         'ALIGN|AXES|AXIS|BGCOLOR|CHAR|CHAROFF|CLASS|COLSPAN|DIR|ID|'.
               'LANG|NOWRAP|ROWSPAN|STYLE|VALIGN',
 'THEAD',      'ALIGN|CHAR|CHAROFF|CLASS|DIR|ID|LANG|STYLE|VALIGN',
 'TITLE',      'DIR|LANG',
 'TR',         'ALIGN|CHAR|CHAROFF|CLASS|DIR|ID|LANG|STYLE|VALIGN',
 'UL',         'CLASS|DIR|ID|LANG|STYLE|TYPE',
);

#========================================================================
# Microsoft elements and attributes
#========================================================================

$msElements = 'APPLET|AREA|BASEFONT|BGSOUND|BIG|CAPTION|CENTER|COL|COLGROUP|'.
              'COMMENT|DIV|EMBED|FONT|FRAME|FRAMESET|IFRAME|MAP|MARQUEE|NOBR|'.
              'NOEMBED|NOFRAMES|NOSCRIPT|OBJECT|PARAM|S|SCRIPT|SMALL|SPAN|'.
              'STRIKE|STYLE|STRIKE|SUB|SUP|TABLE|TBODY|TD|TEXTFLOW|TFOOT|TH|'.
              'THEAD|TR|U|WBR';

%msAttributes =
(
 'A',        'CLASS|ID|LANG|ONCLICK|ONMOUSEOVER|STYLE|TARGET',
 'ADDRESS',  'CLASS|ID|LANG|STYLE',
 'APPLET',   'ALIGN|ALT|BORDER|CLASS|CODE|CODEBASE|DOWNLOAD|HEIGHT|HSPACE|ID|'.
             'LANG|NAME|STYLE|TITLE|VSPACE|WIDTH',
 'AREA',     'ALT|CLASS|COORDS|HREF|ID|LANG|NOHREF|NOTAB|SHAPE|STYLE|'.
             'TABINDEX|TARGET|TITLE',
 'BASE',     'TARGET',
 'BASEFONT', 'COLOR|FONT|NAME|SIZE',
 'BGSOUND',  'CLASS|ID|LANG|LOOP|SRC|START|STYLE',
 'BLOCKQUOTE', 'CLASS|ID|LANG|STYLE',
 'BODY',     'ALINK|BACKGROUND|BGCOLOR|BGPROPERTIES|CLASS|ID|LANG|LEFTMARGIN|'.
             'LINK|STYLE|TEXT|TOPMARGIN|VLINK',
 'BR',       'CLASS|CLEAR|ID|LANG|STYLE',
 'CAPTION',  'ALIGN|CLASS|HEIGHT|ID|LANG|STYLE|VALIGN|WIDTH',
 'CENTER',   'CLASS|ID|LANG|STYLE',
 'COL',      'ALIGN|CHAR|CHAROFF|CLASS|ID|LANG|SPAN|STYLE|VALIGN|WIDTH',
 'COLGROUP', 'ALIGN|CHAR|CHAROFF|CLASS|HALIGN|ID|LANG|SPAN|STYLE|VALIGN|WIDTH',
 'DIR',      'CLASS|COMPACT|ID|LANG|STYLE',
 'DIV',      'ALIGN|CLASS|ID|LANG|NOWRAP|STYLE',
 'DD',       'CLASS|ID|LANG|STYLE',
 'DL',       'CLASS|ID|LANG|STYLE',
 'DT',       'CLASS|ID|LANG|STYLE',
 'EMBED',    'CLASS|HEIGHT|ID|LANG|NAME|OPTIONAL PARAMETER|PALETTE|SRC|'.
             'STYLE|UNITS|WIDTH',
 'FONT',     'CLASS|COLOR|FACE|ID|LANG|SIZE|STYLE',
 'FORM',     'CLASS|ID|LANG|NAME|ONSUBMIT|STYLE|TARGET',
 'FRAME',    'ALIGN|CLASS|FRAMEBORDER|FRAMESPACING|HEIGHT|HSPACE|ID|LANG|'.
             'MARGINHEIGHT|MARGINWIDTH|NAME|NORESIZE|SCROLLING|SRC|STYLE|'.
             'VSPACE|WIDTH',
 'FRAMESET', 'CLASS|COLS|FRAMEBORDER|FRAMESPACING|ID|LANG|LAYOUT|ROWS|STYLE',
 'H1',       'ALIGN|CLASS|ID|LANG|STYLE',
 'H2',       'ALIGN|CLASS|ID|LANG|STYLE',
 'H3',       'ALIGN|CLASS|ID|LANG|STYLE',
 'H4',       'ALIGN|CLASS|ID|LANG|STYLE',
 'H5',       'ALIGN|CLASS|ID|LANG|STYLE',
 'H6',       'ALIGN|CLASS|ID|LANG|STYLE',
 'HR',       'ALIGN|CLASS|COLOR|ID|LANG|NOSHADE|SIZE|STYLE|WIDTH',
 'HTML',     'CLASS|ID|LANG|STYLE',
 'IFRAME',   'ALIGN|BORDER|CLASS|FRAMEBORDER|FRAMESPACING|HEIGHT|ID|LANG|'.
             'MARGINHEIGHT|MARGINWIDTH|NAME|NORESIZE|SCROLLING|SRC|STYLE|'.
             'WIDTH',
 'IMG',      'BORDER|CLASS|CONTROLS|DYNSRC|HEIGHT|HSPACE|ID|LANG|LOOP|START|'.
             'STYLE|USEMAP|VRML|VSPACE|WIDTH',
 'INPUT',    'CLASS|ID|LANG|NOTAB|ONBLUR|ONCHANGE|ONCLICK|ONFOCUS|ONSELECT|'.
             'STYLE|TABINDEX|TITLE',
 'ISINDEX',  'ACTION|CLASS|ID|LANG|PROMPT|STYLE',
 'LI',       'CLASS|ID|LANG|STYLE|TYPE|VALUE',
 'LINK',     'CLASS|ID|LANG|STYLE|TYPE',
 'MAP',      'NAME',
 'MARQUEE',  'ALIGN|BACKGROUND|BEHAVIOR|BGCOLOR|BORDER|CLASS|DIRECTION|'.
             'HEIGHT|HSPACE|ID|LANG|LOOP|SCROLLAMOUNT|SCROLLDELAY|STYLE|'.
             'VSPACE|WIDTH',
 'MENU',     'CLASS|COMPACT|ID|LANG|STYLE',
 'META',     'CHARSET|URL',
 'NOBR',     'CLASS|ID|LANG|STYLE',
 'NOSCRIPT', 'CLASS|ID|LANG|STYLE',
 'OBJECT',   'ALIGN|BORDER|CLASS|CLASSID|CODEBASE|CODETYPE|DATA|DECLARE|'.
             'HEIGHT|HSPACE|ID|ISMAP|LANG|NAME|NOTAB|SHAPES|STANDBY|STYLE|'.
             'TABINDEX|TITLE|TYPE|USEMAP|VSPACE|WIDTH|',
 'OL',       'CLASS|ID|LANG|START|STYLE|TYPE',
 'OPTION',   'CLASS|ID|LANG|STYLE',
 'P',        'ALIGN|CLASS|ID|LANG|STYLE',
 'PARAM',    'NAME|TYPE|VALUE|VALUETYPE',
 'PRE',      'CLASS|ID|LANG|STYLE',
 'SCRIPT',   'CLASS|ID|LANG|LANGUAGE|NAME|SRC|STYLE|TYPE',
 'SELECT',   'CLASS|ID|LANG|STYLE',
 'STYLE',    'TITLE|TYPE',
 'TABLE',    'ALIGN|BACKGROUND|BGCOLOR|BORDER|BORDERCOLOR|BORDERCOLORDARK|'.
             'BORDERCOLORLIGHT|CELLPADDING|CELLSPACING|CLASS|CLEAR|COLS|'.
             'FRAME|ID|LANG|NOWRAP|RULES|STYLE|VALIGN|WIDTH',
 'TBODY',    'ALIGN|CHAR|CHAROFF|CLASS|ID|LANG|STYLE|VALIGN',
 'TD',       'ALIGN|AXES|AXIS|BACKGROUND|BGCOLOR|BORDERCOLOR|BORDERCOLORDARK|'.
             'BORDERCOLORLIGHT|CHAR|CHAROFF|CLASS|COLSPAN|HEIGHT|ID|LANG|'.
             'NOWRAP|ROWSPAN|STYLE|VALIGN|WIDTH',
 'TEXTAREA', 'CLASS|ID|LANG|STYLE',
 'TFOOT',    'ALIGN|CHAR|CHAROFF|CLASS|ID|LANG|STYLE|VALIGN',
 'TH',       'ALIGN|BACKGROUND|BGCOLOR|BORDERCOLOR|BORDERCOLORDARK|'.
             'BORDERCOLORLIGHT|COLSPAN|HEIGHT|NOWRAP|ROWSPAN|VALIGN|WIDTH',
 'THEAD',    'ALIGN|CHAR|CHAROFF|CLASS|ID|LANG|STYLE|VALIGN',
 'TR',       'ALIGN|AXES|AXIS|BACKGROUND|BGCOLOR|BORDERCOLOR|BORDERCOLORDARK|'.
             'BORDERCOLORLIGHT|CHAR|CHAROFF|CLASS|COLSPAN|HEIGHT|ID|LANG|'.
             'NOWRAP|ROWSPAN|STYLE|VALIGN|WIDTH',
 'UL',       'CLASS|ID|LANG|STYLE|TYPE',
 'WBR',      'CLASS|ID|LANG|STYLE',
);

#========================================================================
# Netscape elements and attributes
#========================================================================

$netscapeElements = 'APPLET|AREA|BASEFONT|BIG|BLINK|CAPTION|CENTER|DIV|EMBED|'.
                    'FONT|FRAME|FRAMESET|KEYGEN|MAP|MULTICOL|NOBR|NOEMBED|'.
                    'NOFRAMES|NOSCRIPT|PARAM|S|SCRIPT|SERVER|SMALL|SPACER|'.
                    'STRIKE|SUB|SUP|TABLE|TD|TH|TR|U|WBR';

%netscapeAttributes =
(
 'A',        'ONCLICK|ONMOUSEOUT|ONMOUSEOVER|TARGET',
 'APPLET',   'ALIGN|ALT|ARCHIVE|CODE|CODEBASE|HEIGHT|HSPACE|MAYSCRIPT|NAME|'.
             'VSPACE|WIDTH',
 'AREA',     'COORDS|HREF|NOHREF|ONMOUSEOUT|ONMOUSEOVER|SHAPE|TARGET',
 'BASE',     'TARGET',
 'BASEFONT', 'SIZE',
 'BODY',     'ALINK|BACKGROUND|BGCOLOR|LINK|ONBLUR|ONFOCUS|ONLOAD|ONUNLOAD|'.
             'TEXT|VLINK',
 'BR',       'CLEAR',
 'CAPTION',  'ALIGN',
 'DIV',      'ALIGN',
 'EMBED',    'ALIGN|BORDER|FRAMEBORDER|HEIGHT|HIDDEN|NAME|PALETTE|'.
             'PLUGINSPAGE|SRC|TYPE|VSPACE|WIDTH|'.
# these attributes are for LiveAudio
             'AUTOSTART|CONTROLS|ENDTIME|LOOP|MASTERSOUND|STARTTIME|VOLUME|'.
# these attributes are for QuickTime
             'AUTOPLAY|CONTROLLER|CORRECTION|FOV|HREF|NODE|PAN|'.
             'PLAYEVERYFRAME|TARGET|TILT|'.
# these attributes are for ShockWave
             '|TEXTFOCUS',
 'FONT',     'COLOR|FACE|SIZE',
 'FORM',     'NAME|ONRESET|ONSUBMIT|TARGET',
 'FRAME',    'BORDERCOLOR|FRAMEBORDER|MARGINHEIGHT|MARGINWIDTH|NAME|NORESIZE|'.
             'SCROLLING|SRC',
 'FRAMESET', 'BORDER|BORDERCOLOR|COLS|FRAMEBORDER|ONBLUR|ONFOCUS|ONLOAD|'.
             'ONUNLOAD|ROWS',
 'H1',       'ALIGN',
 'H2',       'ALIGN',
 'H3',       'ALIGN',
 'H4',       'ALIGN',
 'H5',       'ALIGN',
 'H6',       'ALIGN',
 'HR',       'ALIGN|NOSHADE|SIZE|WIDTH',
 'IMG',      'BORDER|HEIGHT|HSPACE|LOWSRC|ONABORT|ONERROR|ONLOAD|USEMAP|'.
             'VSPACE|WIDTH',
 'INPUT',    'ONBLUR|ONCHANGE|ONCLICK|ONFOCUS|ONSELECT',
 'ISINDEX',  'PROMPT',
 'KEYGEN',   'CHALLENGE|NAME',
 'LI',       'TYPE|VALUE',
 'MAP',      'NAME',
 'MULTICOL', 'COLS|GUTTER|WIDTH',
 'OL',       'START|TYPE',
 'P',        'ALIGN',
 'PARAM',    'NAME|VALUE',
 'SCRIPT',   'LANGUAGE|SRC',
 'SELECT',   'ONBLUR|ONCHANGE|ONCLICK|ONFOCUS|SIZE',
 'SPACER',   'ALIGN|HEIGHT|SIZE|TYPE|WIDTH',
 'TABLE',    'ALIGN|BGCOLOR|BORDER|CELLPADDING|CELLSPACING|HEIGHT|HSPACE|'.
             'VSPACE|WIDTH',
 'TD',       'ALIGN|BGCOLOR|COLSPAN|NOWRAP|ROWSPAN|VALIGN',
 'TEXTAREA', 'ONBLUR|ONCHANGE|ONFOCUS|ONSELECT|WRAP',
 'TH',       'ALIGN|BGCOLOR|COLSPAN|NOWRAP|ROWSPAN|VALIGN',
 'TR',       'ALIGN|BGCOLOR|VALIGN',
 'UL',       'TYPE',
);

#========================================================================
# define extension elements
#========================================================================

$extensionElements = "$WilburElements|$i18nElements|$CougarElements|".
                     "$msElements|$netscapeElements";

#========================================================================
# Function: AddExtension
# Purpose:  Extend the HTML understood.
#========================================================================
sub AddExtension
{
   local($extension) = @_;
   local(@extlist);
   local($element);

   #---------------------------------------------------------------------

   if ($extension =~ /,/o)
   {
      @extlist = split(/\s*,\s*/, $extension);
      &AddExtension(shift @extlist) while @extlist > 0;
      return;
   }

   if ($extension !~ /(html)?(3\.2|2\.x)|wilbur|i18n|cougar|microsoft|msie|netscape|mozilla/ && $extension ne '')

   {
      if ($PROGRAM =~ /jweblint/io)
      {
         warn "$PROGRAM: `$extension' ʳĥǤ -- ̵뤷ޤ.\n";
      }
      else
      {
         warn "$PROGRAM: unknown extension `$extension' -- ignoring.\n";
      }
      return;
   }

   #---------------------------------------------------------------------
   # HTML 3.2 (Wilbur) extensions
   #---------------------------------------------------------------------

   if ($extension =~ /(html)?3\.2|wilbur/)
   {
      $legalElements .= '|'.$WilburElements;
      $maybePaired   .= '|TD|TH|TR';
      $pairElements  .= '|APPLET|BIG|CAPTION|CENTER|DIV|FONT|MAP|'.
                        'SCRIPT|SMALL|STRIKE|STYLE|SUB|SUP|TABLE|TD|TH|TR|U';
      $expectArgsRE  .= '|BASEFONT|FONT';

      if ($PROGRAM =~ /jweblint/io)
      {
         $badTextContext{'TABLE'} = 'TD 뤤 TH';
         $badTextContext{'TR'}    = 'TD 뤤 TH';
      }
      else
      {
         $badTextContext{'TABLE'} = 'TD or TH';
         $badTextContext{'TR'}    = 'TD or TH';
      }

      foreach $element (keys %WilburAttributes)
      {
         &AddAttributes($element, split(/\|/, $WilburAttributes{$element}));
      }

      $requiredContext{'AREA'}     = 'MAP';
      $requiredContext{'CAPTION'}  = 'TABLE';
      $requiredContext{'PARAM'}    = 'APPLET';
      $requiredContext{'TD'}       = 'TR';
      $requiredContext{'TH'}       = 'TR';
      $requiredContext{'TR'}       = 'TABLE';

      $requiredAttributes{'AREA'}   = 'ALT';
      $requiredAttributes{'APPLET'} = 'CODE|HEIGHT|WIDTH';
      $requiredAttributes{'PARAM'}  = 'NAME';

      #-- formats for new attributes ---------------------------------------

      $attributeFormat{'BORDER'}  = '\d*';
      $attributeFormat{'CLEAR'}   = 'ALL|LEFT|NONE|RIGHT';
      $attributeFormat{'COLSPAN'} = '\d+';
      $attributeFormat{'HEIGHT'}  = '\d+';
      $attributeFormat{'HSPACE'}  = '\d+';
      $attributeFormat{'ROWSPAN'} = '\d+';
      $attributeFormat{'SHAPE'}   = 'CIRCLE|POLY|RECT';
      $attributeFormat{'START'}   = '\d+';
      $attributeFormat{'VALIGN'}  = 'BOTTOM|MIDDLE|TOP';
      $attributeFormat{'VSPACE'}  = '\d+';
      $attributeFormat{'WIDTH'}   = '\d+%?';

      #-- added formats for existing attributes ----------------------------

      $attributeFormat{'ALIGN'} .= '|CENTER|LEFT|RIGHT';
      $attributeFormat{'SIZE'}  .= '|[-+]?\d+';
      $attributeFormat{'TYPE'}  .= '|[AaIi1]|disc|square|circle|FILE';

      # attribute format check for attributes which take colors
      $attributeFormat{'ALINK'}   = $attributeFormat{'VLINK'} =
      $attributeFormat{'LINK'}    = $attributeFormat{'TEXT'}  =
      $attributeFormat{'BGCOLOR'} =
      $attributeFormat{'COLOR'}   = $colorRE.'|'.$ColorNames;

      $okInHead{'SCRIPT'} = 1;
      $okInHead{'STYLE'}  = 1;

      # ACTION is not required in HTML 3.2 DTD, therefore
      delete $requiredAttributes{'FORM'};
   }

   #---------------------------------------------------------------------
   # i18n extensions
   #---------------------------------------------------------------------

   if ($extension =~ /i18n|(html)?2\.x/)
   {
      foreach $element (keys %i18nAttributes)
      {
         &AddAttributes($element, split(/\|/, $i18nAttributes{$element}));
      }
      $entities      .= '|'.$i18nEntities;
      $legalElements .= '|'.$i18nElements;
      $pairElements  .= '|BDO|Q|SPAN|SUB|SUP';
      $expectArgsRE  .= '|SPAN';

      @commonElements = ('B', 'CITE', 'CODE', 'DFN', 'EM', 'I', 'KBD', 'SAMP',
                         'STRONG', 'TT', 'SUB', 'SUP', 'VAR',
                         'BODY', 'DD', 'DT', 'HEAD', 'ISINDEX', 'LISTING',
                         'PLAINTEXT', 'Q', 'SPAN', 'TITLE', 'XMP');

      foreach $elt (@commonElements)
      {
        $validAttributes{$elt} = 'CLASS|DIR|ID|LANG';
      }

      $requiredAttributes{'BDO'}  = 'DIR';

      $attributeFormat{'DIR'}    = 'LTR|RTL';
      $attributeFormat{'LANG'}   = '[A-Za-z]+(-[A-Za-z]+)*';
      $attributeFormat{'ALIGN'} .= '|CENTER|JUSTIFY|LEFT|RIGHT';
      $attributeFormat{'TYPE'}  .= '|FILE';
   }

   #---------------------------------------------------------------------
   # Cougar extensions
   #---------------------------------------------------------------------

   if ($extension =~ /cougar/)
   {
      $entities      .= '|'.$CougarEntities;
      $legalElements .= '|'.$CougarElements;
      $maybePaired   .= '|TBODY|TD|TFOOT|TH|THEAD|TR';
      $pairElements  .= '|APPLET|BDO|BIG|CAPTION|CENTER|DIV|FIELDSET|FONT|'.
                        'LABEL|MAP|OBJECT|Q|S|SCRIPT|SMALL|SPAN|STYLE|SUB|'.
                        'SUP|TABLE|TBODY|TD|TFOOT|TH|THEAD|TR|U';
      $expectArgsRE  .= '|FONT|SPAN';
      $nonNest       .= '|LABEL';

      if ($PROGRAM =~ /jweblint/io)
      {
         $badTextContext{'COL'}      = 'TD 뤤 TH';
         $badTextContext{'COLGROUP'} = 'TD 뤤 TH';
         $badTextContext{'TABLE'}    = 'TD 뤤 TH';
         $badTextContext{'TBODY'}    = 'TD 뤤 TH';
         $badTextContext{'TFOOT'}    = 'TD 뤤 TH';
         $badTextContext{'THEAD'}    = 'TD 뤤 TH';
         $badTextContext{'TR'}       = 'TD 뤤 TH';
      }
      else
      {
         $badTextContext{'COL'}      = 'TD or TH';
         $badTextContext{'COLGROUP'} = 'TD or TH';
         $badTextContext{'TABLE'}    = 'TD or TH';
         $badTextContext{'TBODY'}    = 'TD or TH';
         $badTextContext{'TFOOT'}    = 'TD or TH';
         $badTextContext{'THEAD'}    = 'TD or TH';
         $badTextContext{'TR'}       = 'TD or TH';
      }

      foreach $element (keys %CougarAttributes)
      {
         &AddAttributes($element, split(/\|/, $CougarAttributes{$element}));
      }

      @fontElements = ('B', 'BIG', 'CITE', 'CODE', 'DFN', 'EM', 'I', 'KBD',
                         'Q', 'S', 'SAMP', 'SMALL', 'STRONG', 'TT', 'SUB',
                         'SUP', 'U', 'VAR');

      foreach $elt (@fontElements)
      {
        $validAttributes{$elt} = 'CLASS|DIR|ID|LANG|STYLE';
      }

      $requiredContext{'AREA'}     = 'MAP';
      $requiredContext{'CAPTION'}  = 'FIELDSET|TABLE';
      $requiredContext{'COL'}      = 'COLGROUP|TABLE';
      $requiredContext{'COLGROUP'} = 'TABLE';
      $requiredContext{'LABEL'}    = 'FORM';
      $requiredContext{'PARAM'}    = 'APPLET|OBJECT';
      $requiredContext{'TBODY'}    = 'TABLE';
      $requiredContext{'TD'}       = 'TR';
      $requiredContext{'TEXTFLOW'} = 'APPLET';
      $requiredContext{'TFOOT'}    = 'TABLE';
      $requiredContext{'TH'}       = 'TR';
      $requiredContext{'THEAD'}    = 'TABLE';
      $requiredContext{'TR'}       = 'TABLE|TBODY|TFOOT|THEAD';

      $requiredAttributes{'AREA'}     = 'ALT';
      $requiredAttributes{'APPLET'}   = 'CODE|HEIGHT|WIDTH';
      $requiredAttributes{'BASEFONT'} = 'SIZE';
      $requiredAttributes{'BDO'}      = 'DIR';
      $requiredAttributes{'PARAM'}    = 'NAME';
      $requiredAttributes{'STYLE'}    = 'TYPE';

      #-- formats for new attributes ---------------------------------------

      $attributeFormat{'BORDER'}    = '\d*%?';
      $attributeFormat{'CLEAR'}     = 'ALL|LEFT|NONE|RIGHT';
      $attributeFormat{'COLSPAN'}   = '\d+';
      $attributeFormat{'DIR'}       = 'LTR|RTL';
      $attributeFormat{'DOWNLOAD'}  = '\d+';
      $attributeFormat{'FRAME'}     = 'above|below|border|box|hsides|lhs|rhs|'.
                                      'void|vsides';
      $attributeFormat{'HEIGHT'}    = '\d+%?';
      $attributeFormat{'HSPACE'}    = '\d+%?';
      $attributeFormat{'LANG'}      = '[A-Za-z]+(-[A-Za-z]+)*';
      $attributeFormat{'ROWSPAN'}   = '\d+';
      $attributeFormat{'RULES'}     = 'ALL|COLS|GROUPS|NONE|ROWS';
      $attributeFormat{'SHAPE'}     = 'CIRCLE|DEFAULT|POLY|RECT';
      $attributeFormat{'SPAN'}      = '\d+';
      $attributeFormat{'START'}     = '\d+';
      $attributeFormat{'TABINDEX'}  = '\d+';
      $attributeFormat{'VALIGN'}    = 'BASELINE|BOTTOM|MIDDLE|TOP';
      $attributeFormat{'VALUETYPE'} = 'DATA|OBJECT|REF';
      $attributeFormat{'VSPACE'}    = '\d+';
      $attributeFormat{'WIDTH'}     = '\d+%?';

      #-- added formats for existing attributes ----------------------------

      $attributeFormat{'ALIGN'} .= '|CENTER|JUSTIFY|LEFT|RIGHT|'.'CHAR|'.
                                   'BASELINE|TEXTBOTTOM|TEXTMIDDLE|TEXTTOP|';
      $attributeFormat{'SIZE'}  .= '|[-+]?\d+';

      # attribute format check for attributes which take colors
      $attributeFormat{'ALINK'}   = $attributeFormat{'VLINK'} =
      $attributeFormat{'LINK'}    = $attributeFormat{'TEXT'}  =
      $attributeFormat{'BGCOLOR'} =
      $attributeFormat{'COLOR'}   = $colorRE.'|'.$ColorNames;

      $okInHead{'SCRIPT'} = 1;
      $okInHead{'STYLE'}  = 1;

      # TYPE attribute of LINK, OBJECT, PARAM, STYLE, SCRIPT can take
      # some kind of Internet media type, therefore
      delete $attributeFormat{'TYPE'};
   }

   #---------------------------------------------------------------------
   # Microsoft extensions
   #---------------------------------------------------------------------
   if ($extension =~ /microsoft|msie/)
   {
      $legalElements .= '|'.$msElements;
      $maybePaired   .= '|IFRAME|MARQUEE|TBODY|TD|TEXTFLOW|TFOOT|TH|'.
                        'THEAD|TR';
      $pairElements  .= '|APPLET|BIG|CAPTION|CENTER|COMMENT|DIV|EMBED|FONT|'.
                        'FRAMESET|IFRAME|MAP|MARQUEE|NOBR|NOEMBED|NOFRAMES|'.
                        'NOSCRIPT|OBJECT|S|SCRIPT|SMALL|SPAN|STRIKE|'.
                        'STYLE|SUB|SUP|TABLE|TBODY|TD|TEXTFLOW|TFOOT|TH|'.
                        'THEAD|TR|U';
      $expectArgsRE  .= '|BASEFONT|FONT';

      @fontElements   = ('B', 'BIG', 'I', 'SMALL', 'STRIKE', 'STRONG', 'SUB',
                         'SUP', 'TT', 'U');

      foreach $elt (@fontElements)
      {
        $validAttributes{$elt} = 'CLASS|COLOR|FACE|ID|LANG|SIZE|STYLE';
      }

      @phraseElements = ('CITE', 'CODE', 'DFN', 'EM', 'KBD', 'S', 'SAMP',
                         'SPAN', 'STRONG', 'VAR',
                         'XMP', 'LISTING', 'PLAINTEXT');

      foreach $elt (@phraseElements)
      {
        $validAttributes{$elt} = 'CLASS|ID|LANG|STYLE';
      }

      $mustFollow{'FRAMESET'} = '/HEAD|FRAME|FRAMESET|/FRAMESET|/NOFRAMES';
      $mustFollow{'/HTML'}    = '/BODY|/FRAMESET';

      if ($PROGRAM =~ /jweblint/io)
      {
         $badTextContext{'COL'}      = 'TD 뤤 TH';
         $badTextContext{'COLGROUP'} = 'TD 뤤 TH';
         $badTextContext{'TABLE'}    = 'TD 뤤 TH';
         $badTextContext{'TBODY'}    = 'TD 뤤 TH';
         $badTextContext{'TFOOT'}    = 'TD 뤤 TH';
         $badTextContext{'THEAD'}    = 'TD 뤤 TH';
         $badTextContext{'TR'}       = 'TD 뤤 TH';
      }
      else
      {
         $badTextContext{'COL'}      = 'TD or TH';
         $badTextContext{'COLGROUP'} = 'TD or TH';
         $badTextContext{'TABLE'}    = 'TD or TH';
         $badTextContext{'TBODY'}    = 'TD or TH';
         $badTextContext{'TFOOT'}    = 'TD or TH';
         $badTextContext{'THEAD'}    = 'TD or TH';
         $badTextContext{'TR'}       = 'TD or TH';
      }
      $badTextContext{'FRAMESET'} = 'NOFRAMES';

      foreach $element (keys %msAttributes)
      {
         &AddAttributes($element, split(/\|/, $msAttributes{$element}));
      }

      $requiredContext{'AREA'}     = 'MAP|OBJECT';
      $requiredContext{'CAPTION'}  = 'TABLE';
      $requiredContext{'COL'}      = 'COLGROUP|TABLE';
      $requiredContext{'COLGROUP'} = 'TABLE';
      $requiredContext{'FRAME'}    = 'FRAMESET';
      $requiredContext{'NOEMBED'}  = 'EMBED';
      $requiredContext{'NOFRAMES'} = 'FRAMESET';
      $requiredContext{'PARAM'}    = 'APPLET|OBJECT';
      $requiredContext{'TBODY'}    = 'TABLE';
      $requiredContext{'TD'}       = 'TR';
      $requiredContext{'TFOOT'}    = 'TABLE';
      $requiredContext{'TH'}       = 'TR';
      $requiredContext{'THEAD'}    = 'TABLE';
      $requiredContext{'TR'}       = 'TABLE|TBODY|TFOOT|THEAD';

      $requiredAttributes{'APPLET'}   = 'CODE|HEIGHT|WIDTH';
      $requiredAttributes{'AREA'}     = 'COORDS';
      $requiredAttributes{'BGSOUND'}  = 'SRC';
      $requiredAttributes{'EMBED'}    = 'SRC';
      $requiredAttributes{'FRAME'}    = 'SRC';
      $requiredAttributes{'IFRAME'}   = 'SRC';
      $requiredAttributes{'PARAM'}    = 'NAME';

      #-- formats for new attributes ---------------------------------------

      #-- FRAME attributes
      $attributeFormat{'FRAMEBORDER'}  = 'NO|YES|[01]';
      $attributeFormat{'FRAMESPACING'} = '\d+';
      $attributeFormat{'MARGINHEIGHT'} = '\d+';
      $attributeFormat{'MARGINWIDTH'}  = '\d+';
      $attributeFormat{'SCROLLING'}    = 'AUTO|NO|YES';

      # this should really be specific to ROWS and COLS in FRAMESET element
      $attributeFormat{'ROWS'} =
      $attributeFormat{'COLS'} = '\d+|(\d*[*%]?,)*\d*[*%]?';

      #-- IMG attributes
      $attributeFormat{'BORDER'}       = '\d*';
      $attributeFormat{'START'}   = '(FILEOPEN)|(MOUSEOVER)|\1,\s*\2|\2,\s*\1';

      #-- MARQUEE attributes
      $attributeFormat{'BEHAVIOR'}     = 'ALTERNATE|SCROLL|SLIDE';
      $attributeFormat{'DIRECTION'}    = 'LEFT|RIGHT';
      $attributeFormat{'HEIGHT'}       = '\d+%?';
      $attributeFormat{'HSPACE'}       = '\d+';
      $attributeFormat{'SCROLLAMOUNT'} = '\d+';
      $attributeFormat{'SCROLLDELAY'}  = '\d+';
      $attributeFormat{'VSPACE'}       = '\d+';
      $attributeFormat{'WIDTH'}        = '\d+%?';

      #-- TABLE attributes
      $attributeFormat{'FRAME'}        = 'ABOVE|BELOW|BORDER|BOX|HSIDES|'.
                                         'LHS|RHS|VOID|VSIDES';
      $attributeFormat{'RULES'}        = 'ALL|COLS|GROUPS|NONE|ROWS';
      $attributeFormat{'COLSPAN'}      = '\d+';
      $attributeFormat{'ROWSPAN'}      = '\d+';

      #-- COL|COLGROUP attributes
      $attributeFormat{'SPAN'}         = '\d+';

      $attributeFormat{'BGPROPERTIES'} = 'FIXED';
      $attributeFormat{'CLEAR'}        = 'LEFT|RIGHT|ALL';
      $attributeFormat{'LANG'}         = '[A-Za-z]+(-[A-Za-z]+)*';
      $attributeFormat{'LEFTMARGIN'}   = '\d+';
      $attributeFormat{'LOOP'}         = '\d+|INFINITE';
      # for EMBED
      $attributeFormat{'PALETTE'}      = '\#[0-9a-fA-F]{6}\|\#[0-9a-fA-F]{6}';
      $attributeFormat{'SHAPE'}        = 'RECT|RECTANGLE|CIRC|CIRCLE|POLY|'.
                                         'POLYGON';	# for AREA
      $attributeFormat{'TABINDEX'}     = '\d+';
      $attributeFormat{'TOPMARGIN'}    = '\d+';
      $attributeFormat{'UNITS'}        = 'EN|PIXELS';	# for EMBED
      $attributeFormat{'VALIGN'}       = 'BASELINE|BOTTOM|MIDDLE|TOP';
      $attributeFormat{'VALUETYPE'}    = 'DATA|OBJECT|REF';	# for PARAM

      #-- added formats for existing attributes ----------------------------

      $attributeFormat{'ALIGN'} .= '|LEFT|CENTER|RIGHT|JUSTIFY|'.
                                   # for OBJECT
                                   'BASELINE|TEXTBOTTOM|TEXTMIDDLE|TEXTTOP';
      $attributeFormat{'SIZE'}  .= '|[-+]?\d+';

      # attribute format check for attributes which take colors
      $attributeFormat{'ALINK'}   = $attributeFormat{'VLINK'} =
      $attributeFormat{'LINK'}    = $attributeFormat{'TEXT'}  =
      $attributeFormat{'BGCOLOR'} =
      $attributeFormat{'COLOR'}   = $attributeFormat{'BORDERCOLOR'} =
      $attributeFormat{'BORDERCOLORLIGHT'} =
      $attributeFormat{'BORDERCOLORDARK'}  = $colorRE.'|'.$ColorNames;

      $okInHead{'BGSOUND'} = 1;
      $okInHead{'SCRIPT'}  = 1;
      $okInHead{'STYLE'}   = 1;

      # IE 3.0 DTD requires ACTION attribute for FORM element,
      # but actually it's not required when script is used, therefore
      delete $requiredAttributes{'FORM'};

      # TYPE attribute of OBJECT and PARAM elements can take
      # some kind of Internet media type, therefore
      delete $attributeFormat{'TYPE'};
   }

   #---------------------------------------------------------------------
   # Netscape extensions
   #---------------------------------------------------------------------

   if ($extension =~ /netscape|mozilla/)
   {
      $legalElements .= '|'.$netscapeElements;
      $maybePaired   .= '|EMBED|TD|TH|TR';
      $pairElements  .= '|APPLET|BIG|BLINK|CAPTION|CENTER|DIV|EMBED|FONT|'.
                        'FRAMESET|MULTICOL|NOBR|NOEMBED|NOFRAMES|NOSCRIPT|'.
                        'MAP|S|SCRIPT|SERVER|SMALL|STRIKE|SUB|SUP|TABLE|TD|'.
                        'TH|TR|U';
      $expectArgsRE  .= '|BASE|EMBED|FONT|FRAMESET|SCRIPT';

      $mustFollow{'FRAMESET'} = '/HEAD|FRAME|FRAMESET|/FRAMESET|/NOFRAMES';
      $mustFollow{'/HTML'}    = '/BODY|/FRAMESET';

      if ($PROGRAM =~ /jweblint/io)
      {
         $badTextContext{'TABLE'}    = 'TD 뤤 TH';
         $badTextContext{'TR'}       = 'TD 뤤 TH';
      }
      else
      {
         $badTextContext{'TABLE'}    = 'TD or TH';
         $badTextContext{'TR'}       = 'TD or TH';
      }
      $badTextContext{'FRAMESET'} = 'NOFRAMES';

      foreach $element (keys %netscapeAttributes)
      {
         &AddAttributes($element, split(/\|/, $netscapeAttributes{$element}));
      }

      $requiredContext{'AREA'}     = 'MAP';
      $requiredContext{'CAPTION'}  = 'TABLE';
      $requiredContext{'FRAME'}    = 'FRAMESET';
      $requiredContext{'KEYGEN'}   = 'FORM';
      $requiredContext{'NOFRAMES'} = 'FRAMESET';
      $requiredContext{'PARAM'}    = 'APPLET';
      $requiredContext{'TD'}       = 'TR';
      $requiredContext{'TH'}       = 'TR';
      $requiredContext{'TR'}       = 'TABLE';

      $requiredAttributes{'AREA'}     = 'COORDS|HREF';
      $requiredAttributes{'APPLET'}   = 'CODE|HEIGHT|WIDTH';
      $requiredAttributes{'BASEFONT'} = 'SIZE';
      $requiredAttributes{'FRAME'}    = 'SRC';
      $requiredAttributes{'KEYGEN'}   = 'NAME';
      $requiredAttributes{'MAP'}      = 'NAME';
      $requiredAttributes{'MULTICOL'} = 'COLS';
      $requiredAttributes{'PARAM'}    = 'NAME|VALUE';

      #-- formats for new attributes ---------------------------------------

      # these are for EMBED
      $attributeFormat{'AUTOPLAY'}       = 'TRUE|FALSE';
      $attributeFormat{'AUTOSTART'}      = 'TRUE|FALSE|PALINDROME';
      $attributeFormat{'CONTROLLER'}     = 'TRUE|FALSE';
      $attributeFormat{'CONTROLS'}       = 'CONSOLE|SMALLCONSOLE|PLAYBUTTON|'.
                                           'STOPBUTTOM|VOLUMELEVER';
      $attributeFormat{'CORRECTION'}     = 'NONE|PARTIAL|FULL';
      $attributeFormat{'ENDTIME'}        = '\d{2}:\d{2}';
      $attributeFormat{'HIDDEN'}         = 'TRUE|FALSE';
      $attributeFormat{'LOOP'}           = 'TRUE|FALSE|INTEGER';
      $attributeFormat{'NODE'}           = '\d+';
      $attributeFormat{'PALETTE'}        = 'FOREGROUND|BACKGROUND';
      $attributeFormat{'PLAYEVERYFRAME'} = 'TRUE|FALSE';
      $attributeFormat{'STARTTIME'}      = '\d{2}:\d{2}';
      $attributeFormat{'TEXTFOCUS'}      = 'onMouse|onStart|never';
      $attributeFormat{'VOLUME'}         = '\d{1,2}|100';

      #-- FRAME attributes
      $attributeFormat{'FRAMEBORDER'}    = 'NO|YES';
      $attributeFormat{'MARGINHEIGHT'}   = '\d+';
      $attributeFormat{'MARGINWIDTH'}    = '\d+';
      $attributeFormat{'SCROLLING'}      = 'AUTO|NO|YES';

      # this should really be specific to ROWS and COLS in FRAMESET element
      $attributeFormat{'ROWS'} =
      $attributeFormat{'COLS'} = '\d+|(\d*[*%]?,)*\d*[*%]?';

      $attributeFormat{'BORDER'}         = '\d*';
      $attributeFormat{'CLEAR'}          = 'ALL|LEFT|RIGHT';
      $attributeFormat{'COLSPAN'}        = '\d+';
      $attributeFormat{'COLUMNS'}        = '\d+';
      $attributeFormat{'GUTTER'}         = '\d+';	# for MULTICOL
      $attributeFormat{'HEIGHT'}         = '\d+%?';
      $attributeFormat{'HSPACE'}         = '\d+';
      $attributeFormat{'NEEDS'}          = '\d+\s*EM|\d+\s*PIXELS';
      $attributeFormat{'ROWSPAN'}        = '\d+';
      $attributeFormat{'SIZE'}           = '[-+]?\d+';
      $attributeFormat{'VALIGN'}         = 'BASELINE|BOTTOM|MIDDLE|TOP';
      $attributeFormat{'VSPACE'}         = '\d+';
      $attributeFormat{'WIDTH'}          = '\d+%?';
      $attributeFormat{'WRAP'}           = 'HARD|OFF|PHYSICAL|SOFT|VIRTUAL';

      #-- added formats for existing attributes ----------------------------

      $attributeFormat{'ALIGN'} .= '|ABSMIDDLE|ABSBOTTOM|BASELINE|'.
                                   'BLEEDLEFT|BLEEDRIGHT|TEXTTOP|'.
                                   'LEFT|CENTER|RIGHT|JUSTIFY';

      # attribute format check for attributes which take colors
      $attributeFormat{'ALINK'} = $attributeFormat{'VLINK'}   =
      $attributeFormat{'LINK'}  = $attributeFormat{'TEXT'}    =
      $attributeFormat{'COLOR'} = $attributeFormat{'BGCOLOR'} =
      $attributeFormat{'BORDERCOLOR'} = $colorRE.'|'.$ColorNames;

      # BASE can take just a TARGET attribute, HREF not required therefore
      delete $requiredAttributes{'BASE'};

      # FORM can take just a NAME attribute, ACTION not required therefore
      delete $requiredAttributes{'FORM'};

      # TYPE attribute of EMBED element can take some kind of 
      # Internet media type, therefore
      delete $attributeFormat{'TYPE'};

      $okInHead{'SCRIPT'}   = 1;
      $okInHead{'NOSCRIPT'} = 1;
   }
}

#========================================================================
# Function: AddAttributes
# Purpose:  Add extended attributes.
#========================================================================

sub AddAttributes
{
   local($element,@attributes) = @_;
   local($attr);


   $attr = join('|', @attributes);
   if (defined $validAttributes{$element})
   {
      $validAttributes{$element} .= "|$attr";
   }
   else
   {
      $validAttributes{$element} = "$attr";
   }
}

#========================================================================
# Function: ExtensionAttributes
# Purpose:  Define extension attributes, which is not included
#           in %validAttributes.
#========================================================================

sub ExtensionAttributes
{
   local($opt) = @_;

   $opt .= ",".$doctype;

   if ($opt !~ /(html)?3\.2|wilbur/)
   {
      foreach $element (keys %WilburAttributes)
      {
         &AddExtensionAttributes($element, 
          split(/\|/, $WilburAttributes{$element}));
      }
   }

   if ($opt !~ /i18n|(html)?2\.x/)
   {
      foreach $element (keys %i18nAttributes)
      {
         &AddExtensionAttributes($element, 
          split(/\|/, $i18nAttributes{$element}));
      }
   }

   if ($opt !~ /cougar/)
   {
      foreach $element (keys %CougarAttributes)
      {
         &AddExtensionAttributes($element, 
          split(/\|/, $CougarAttributes{$element}));
      }
   }

   if ($opt !~ /microsoft|msie/)
   {
      foreach $element (keys %msAttributes)
      {
         &AddExtensionAttributes($element, 
          split(/\|/, $msAttributes{$element}));
      }
   }

   if ($opt !~ /netscape|mozilla/)
   {
      foreach $element (keys %netscapeAttributes)
      {
         &AddExtensionAttributes($element, 
          split(/\|/, $netscapeAttributes{$element}));
      }
   }

   $ExtensionAttributes = 1;
}

#========================================================================
# Function: AddExtensionAttributes
# Purpose:  Add extension attributes (called from &ExtensionAttributes).
#========================================================================

sub AddExtensionAttributes
{
   local($element,@attributes) = @_;
   local($attr);


   $attr = join('|', @attributes);
   if (defined $extensionAttributes{$element})
   {
      $extensionAttributes{$element} .= "|$attr";
   }
   else
   {
      $extensionAttributes{$element} = "$attr";
   }
}

#========================================================================
# Function: CheckDoctype
# Purpose:  Check validity of selected extension in accordance with
#           DOCTYPE declaration.
#========================================================================

sub CheckDoctype
{
   local($id,$extension) = @_;
   local($selectedExtension);

   if ($id =~ m#$WilburDocType#io)
   {
      $doctype = 'wilbur';
   }
   elsif ($id =~ m#$i18nDocType#io)
   {
      $doctype = 'i18n';
   }
   elsif ($id =~ m#$CougarDocType#io)
   {
      $doctype = 'cougar';
   }
   elsif ($id =~ m#$msDocType#io)
   {
      $doctype = 'microsoft';
   }
   elsif ($id =~ /netscape|mozilla/io)
   {
      $doctype = 'netscape';
   }
   else
   {
      $doctype = 'HTML 2.0';
   }

   if ($extension eq undef)
   {
      $selectedExtension = $extension = $doctype;
   }
   else
   {
      $selectedExtension = $extension;
      $selectedExtension =~ s/(HTML)?3\.2/Wilbur/i;
      $selectedExtension =~ s/(HTML)?2\.x/i18n/i;
      $selectedExtension =~ s/MSIE/Microsoft/i;
      $selectedExtension =~ s/Mozilla/Netscape/i;
   }

   if ($selectedExtension !~ /$doctype/io)
   {
      &whine($., 'doctype-mismatch', $extension);
   }
}

#========================================================================
# Function: SetExtension
# Purpose:  Set HTML extensions in accordance with DOCTYPE declaration.
#========================================================================

sub SetExtension
{
   local($id,$ext) = @_;

   if (($id =~ m#$WilburDocType#i) && ($ext !~ /(html)?3\.2|wilbur/i))
   {
      &AddExtension("html3.2");
   }
   elsif (($id =~ m#$i18nDocType#i) && ($ext !~ /i18n|(html)?2\.x/i))
   {
      &AddExtension("i18n");
   }
   elsif (($id =~ m#$CougarDocType#i) && ($ext !~ /cougar/i))
   {
      &AddExtension("cougar");
   }
   elsif (($id =~ m#$msDocType#i) && ($ext !~ /microsoft|msie/i))
   {
      &AddExtension("microsoft");
   }
}

1;
