#========================================================================
# Package:  htmli18n.pl, for j?weblint 97, v0.12 [1997-03-08]
# Purpose:  Handle various kinds of encodings, such as ISO-2022-* and Unicode.
#
# WARNING:  This implementaion is very primitive; it should be refined
#           and enhanced in the future.
#
# [Related Resources]
# RFC 1468: Japanese Character Encoding for Internet Messages [ISO-2022-JP]
# RFC 1554: ISO-2022-JP-2: Multilingual Extension of ISO-2022-JP
# RFC 1557: Korean Character Encoding for Internet Messages [ISO-2022-KR]
# RFC 1922: Chinese Character Encoding for Internet Messages [ISO-2022-CN,
#   ISO-2022-CN-EXT]
# RFC 1642: UTF-7 - A Mail-Safe Transformation Format of Unicode
# RFC 2044: UTF-8, a transformation format of Unicode and ISO 10646
# draft-ohta-text-encoding-01.txt: Internet Multilingual Text Encoding:
#   ISO-2022-INT-* (maybe expired) [ISO-2022-INT-1]
# Mule: MULtilingual Enhancement to GNU Emacs
#========================================================================

# selected registered ISO-2022 escape sequences
%ESC = (
# 'registered number',	"escape sequence",	# character set [source]
'R2',	"\x1b\\x28\x40",	# ISO 646 IRV: 1983
'R4',	"\x1b\\x28\x41",	# BSI 4730 (ISO 646, British Version)
'R6',	"\x1b[\x20\x28]\x42",	# ANS X3.4-1968 (ISO 646, USA Version)
'R13',	"\x1b[\x28\x29]\x49",	# JIS X 0201-1976 ('Katakana' set)
'R14',	"\x1b\\x28\x4a",	# JIS X 0201-1976 ('Roman' set) [JP, JP-2, INT-1]
'R42',	"\x1b\\x24\\x28?\x40",	# JIS X 0208-1978 [JP, JP-2, INT-1]
'R57',	"\x1b\\x24\x54",	# GB 1988-1980
'R58',	"\x1b\\x24[\x28\x29]?\x41",	# GB 2312-1980 [JP-2, CN, CN-EXT, INT-1]
'R87',	"\x1b\\x24\x42",	# JIS X 0208-1983 [JP, JP-2, INT-1]
'R100',	"\x1b[\x2c\x2d\x2e]\x41",	# ISO 8859-1 Latin alphabet No. 1 [JP-2, INT-1]
'R101',	"\x1b[\x2c\x2d\x2e]\x42",	# ISO 8859-2 Latin alphabet No. 2
'R109',	"\x1b[\x2c\x2d\x2e]\x43",	# ISO 8859-3 Latin alphabet No. 3
'R110',	"\x1b[\x2c\x2d\x2e]\x44",	# ISO 8859-4 Latin alphabet No. 4
'R126',	"\x1b[\x2c\x2d\x2e]\x46",	# ISO 8859-7 Latin/Greek alphabet [JP-2]
'R127',	"\x1b[\x2c\x2d\x2e]\x47",	# ISO 8859-6 Latin/Arabic alphabet
'R138',	"\x1b[\x2c\x2d\x2e]\x48",	# ISO 8859-8 Latin/Hebrew alphabet
'R144',	"\x1b[\x2c\x2d\x2e]\x4c",	# ISO 8859-5 Latin/Cyrillic alphabet
'R148',	"\x1b[\x2c\x2d\x2e]\x4d",	# ISO 8859-9 Latin alphabet No. 5
'R149',	"\x1b\\x24(\\x28|\\x29?)\x43",	# KSC 5601-1987 [JP-2, KR, INT-1]
'R157',	"\x1b[\x2c\x2d\x2e]\x56",	# ISO 8859-10 Latin alphabet No. 6
'R159',	"\x1b\\x24\\x28?\x44",	# JIS X 0212-1990 [JP-2, INT-1]
'R165',	"\x1b\\x24\\x29\x45",	# ISO-IR-165 [CN-EXT]
'R166',	"\x1b[\x2c\x2d\x2e]\x54",	# Thai Set
'R171',	"\x1b\\x24\\x28\x47",	# CNS 11643-plane-1 [INT-1]
'R171s',	"\x1b\\x24\\x29\x47",	# CNS 11643-plane-1 [CN, CN-EXT]
'R172',	"\x1b\\x24\\x28\x48",	# CNS 11643-plane-2 [INT-1]
'R172s',	"\x1b\\x24\\x2a\x48",	# CNS 11643-plane-2 [CN, CN-EXT];SS2
'R180',	"\x1b[\x2d\x2e]\x5a",	# VISCII (Vietnamese)
'R183',	"\x1b\\x24\\x2b\x49",	# CNS 11643-plane-3 [CN-EXT];SS3
'R184',	"\x1b\\x24\\x2b\x4a",	# CNS 11643-plane-4 [CN-EXT];SS3
'R185',	"\x1b\\x24\\x2b\x4b",	# CNS 11643-plane-5 [CN-EXT];SS3
'R186',	"\x1b\\x24\\x2b\x4c",	# CNS 11643-plane-6 [CN-EXT];SS3
'R187',	"\x1b\\x24\\x2b\x4d",	# CNS 11643-plane-7 [CN-EXT];SS3
'SI',	"\x0f",	# ISO-2022-CN & ISO-2022-KR & ISO-2022-INT-1 SI
'SO',	"\x0e",	# ISO-2022-CN & ISO-2022-KR & ISO-2022-INT-1 SO
'SI-I',	"\x1b\x0f",	# ISO-2022-INT-1 SI
'SO-I',	"\x1b\x0e",	# ISO-2022-INT-1 SO (indicates KSC 5601)
'SS2',	"\x1b\x4e",	# ISO-2022-CN & ISO-2022-JP-2 SS2
'SS3',	"\x1b\x4f",	# ISO-2022-CN-EXT SS3
'BIG5',	"\x1b\\x24\\x28[\x30\x31]",	# Chenese BIG5 level 1 & 2 [Mule]
'VISCII',	"\x1b[\x2c\x2d\x2e]\x31",	# Vietnamese [Mule]
'Ethio',	"\x1b\\x24\\x28\x32",	# Ethiopian [Mule]
'Arabic0',	"\x1b\\x28\x32",	# Arabic digit [Mule]
'Arabic1',	"\x1b\\x28\x33",	# 1-column Arabic [Mule]
'Arabic2',	"\x1b\\x28\x34",	# 2-column Arabic [Mule]
);

# Unicode characters equivalent to ASCII
$UniASCII = "\x00[\x00-\x7f]";

#========================================================================
# Function: ConvISO2022
# Purpose:  Convert ISO-2022-* coded characters
#             i.e. JIS X 0208-1983 characters are converted to "[R87]"
#========================================================================
sub ConvISO2022
{
   s/$ESC{'R2'}/\x00R2 /g;
   s/$ESC{'R4'}/\x00R4 /g;
   s/$ESC{'R6'}/\x00R6 /g;
   s/$ESC{'R13'}/\x00R13 /g;
   s/$ESC{'R14'}/\x00R14 /g;
   s/$ESC{'R42'}/\x00R42 /g;
   s/$ESC{'R57'}/\x00R57 /g;
   s/$ESC{'R58'}/\x00R58 /g;
   s/$ESC{'R87'}/\x00R87 /g;
   s/$ESC{'R100'}/\x00R100 /g;
   s/$ESC{'R101'}/\x00R101 /g;
   s/$ESC{'R109'}/\x00R109 /g;
   s/$ESC{'R110'}/\x00R110 /g;
   s/$ESC{'R126'}/\x00R126 /g;
   s/$ESC{'R127'}/\x00R127 /g;
   s/$ESC{'R138'}/\x00R138 /g;
   s/$ESC{'R144'}/\x00R144 /g;
   s/$ESC{'R148'}/\x00R148 /g;
   s/$ESC{'R149'}/\x00R149 /g;
   s/$ESC{'R157'}/\x00R157 /g;
   s/$ESC{'R159'}/\x00R159 /g;
   s/$ESC{'R165'}/\x00R165 /g;
   s/$ESC{'R166'}/\x00R166 /g;
   s/$ESC{'R171'}/\x00R171 /g;
   s/$ESC{'R171s'}/[R171]/g;
   s/$ESC{'R172'}/\x00R172 /g;
   s/$ESC{'R180'}/\x00R180 /g;
   s/$ESC{'R183'}/\x00R183 /g;
   s/$ESC{'R184'}/\x00R184 /g;
   s/$ESC{'R185'}/\x00R185 /g;
   s/$ESC{'R186'}/\x00R186 /g;
   s/$ESC{'R187'}/\x00R187 /g;
   s/$ESC{'BIG5'}/\x00BIG5 /g;
   s/$ESC{'VISCII'}/\x00VISCII /g;
   s/$ESC{'Ethio'}/\x00Ethio /g;
   s/$ESC{'Arabic0'}/\x00Arabic0 /g;
   s/$ESC{'Arabic1'}/\x00Arabic1 /g;
   s/$ESC{'Arabic2'}/\x00Arabic2 /g;
   s/$ESC{'SO'}([^\x0e\x0f])*$ESC{'SI'}/\x00[SO]/g;
   s/($ESC{'R172s'})(([^\x1b])*)$ESC{'SS2'}../$1$2[R172]/g;	# ISO-2022-CN SS2
   s/$ESC{'R172s'}/[R172]/g;
   s/$ESC{'SS2'}./[SS2]/g;	# ISO-2022-JP-2 SS2
   s/$ESC{'SS3'}../[SS3]/g;
   s/\x00R([246]|14) /\x00/g;
   s/\x00(R\d+|BIG5|VISCII|Ethio|Arabic[0-2]) [^\x00]*(\x00)/\x00[$1]$2/g;
   s/\x00(R\d+) $/\x00[$1]/g;
   s/\x00//g;
}

#========================================================================
# Function: ConvUnicode
# Purpose:  Convert Unicode.
#========================================================================
sub ConvUnicode
{
   local(@Uni);
   local($UNI);
   local($UniLine);

   @Uni = split(//, $_);
   for (@Uni)
   {
      $UNI .= $_;
      if (length($UNI) == 2)
      {
         if ($UNI =~ $UniASCII)
         {
            $UNI =~ s/\x00([\x00-\x7f])/$1/;
            $UniLine .= $UNI;
         }
         else
         {
            $UniLine .= "[UNI]";
         }
         undef $UNI;
      }
   }
   $UniLine =~ s/(\[UNI\])+/[UNICODE]/g;
   return $UniLine;
}

#========================================================================
# Function: CodeType
# Purpose:  Select the code type to process
#========================================================================
sub CodeType
{
   local($codetype) = @_;

   $CodeType = "Literal"       if ($codetype =~ /literal/);
   $CodeType = "ISO-2022-7bit" if ($codetype =~ /iso-?2022(-7bit)?/);
   $CodeType = "Unicode"       if ($codetype =~ /unicode|ucs-?2/);
   $CodeType = "Japanese"      if ($codetype =~ /japanese|ja|jp/);

   if ($codetype !~ /literal|iso-?2022(-7bit)?|unicode|ucs-?2|japanese|ja|jp/)
   {
      if ($PROGRAM =~ /jweblint/io)
      {
         warn "$PROGRAM: `$codetype' ʥɷϤǤ -- ̵뤷ޤ.\n";
      }
      else
      {
         warn "$PROGRAM: unknown code type `$codetype' -- ignoring.\n";
      }
      return;
   }
}

1;
